"""
domain invariant adversarial trainer hyper-parmaeters
"""


def add_args2parser_dial(parser):
    """
    append hyper-parameters to the main argparser
    """
    parser.add_argument(
        "--dial_steps_perturb",
        type=int,
        default=3,
        help="how many gradient step to go to find an image as adversarials",
    )
    parser.add_argument(
        "--dial_noise_scale",
        type=float,
        default=0.001,
        help="variance of gaussian noise to inject on pure image",
    )
    parser.add_argument(
        "--dial_lr",
        type=float,
        default=0.003,
        help="learning rate to generate adversarial images",
    )
    parser.add_argument(
        "--dial_epsilon",
        type=float,
        default=0.031,
        help="pixel wise threshold to perturb images",
    )
    return parser



import os
from pathlib import Path

import matplotlib.pyplot as plt
from torchvision.utils import make_grid, save_image

from domainlab.utils.logger import Logger


def mk_fun_sav_img(path=".", nrow=8, folder_na=""):
    """
    create torchvision.utils image saver
    """
    Path(path).mkdir(parents=True, exist_ok=True)

    def my_sav_img(comparison_tensor_stack, name, title=None):
        f_p = os.path.join(path, folder_na, name)
        Path(os.path.dirname(f_p)).mkdir(parents=True, exist_ok=True)
        logger = Logger.get_logger()
        logger.info(f"saving to {f_p}")
        # works also if tensor is already in cpu
        tensor = comparison_tensor_stack.cpu()
        if title is None:
            save_image(tensor=tensor, nrow=nrow, fp=f_p)
        else:
            img_grid = make_grid(tensor=tensor, nrow=nrow)
            sav_add_title(img_grid, path=f_p, title="hi")

    return my_sav_img


def sav_add_title(grid_img, path, title):
    """
    add title and save image as matplotlib.pyplot
    """
    fig = plt.gcf()  # get current figure
    plt.imshow(grid_img.permute(1, 2, 0))
    plt.title(title)
    fig.savefig(path)
    fig.clf()  # clear figure



"""
operations that all claasification model should have
"""

import abc
import math

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import functional as F

from domainlab.models.a_model import AModel
from domainlab.utils.logger import Logger
from domainlab.utils.perf import PerfClassif
from domainlab.utils.perf_metrics import PerfMetricClassif
from domainlab.utils.utils_class import store_args
from domainlab.utils.utils_classif import get_label_na, logit2preds_vpic


class AModelClassif(AModel, metaclass=abc.ABCMeta):
    """
    operations that all classification model should have
    """

    match_feat_fun_na = "cal_logit_y"

    def extend(self, model):
        super().extend(model)
        self._net_classifier = model.net_classifier

    @property
    def metric4msel(self):
        return "acc"

    @property
    def net_classifier(self):
        return self._net_classifier

    @net_classifier.setter
    def net_classifier(self, net_classifier):
        self._net_classifier = net_classifier

    def create_perf_obj(self, task):
        """
        for classification, dimension of target can be quieried from task
        """
        self.perf_metric = PerfMetricClassif(task.dim_y)
        return self.perf_metric

    def cal_perf_metric(self, loader, device):
        """
        classification performance metric
        """
        metric = None
        with torch.no_grad():
            if loader is not None:
                metric = self.perf_metric.cal_metrics(self, loader, device)
                confmat = metric.pop("confmat")
                logger = Logger.get_logger()
                logger.info("scalar performance:")
                logger.info(str(metric))
                logger.debug("confusion matrix:")
                logger.debug(pd.DataFrame(confmat))
                metric["confmat"] = confmat
        return metric

    def evaluate(self, loader_te, device):
        """
        for classification task, use the current model to cal acc
        """
        acc = PerfClassif.cal_acc(self, loader_te, device)
        logger = Logger.get_logger()
        logger.info(f"before training, model accuracy: {acc}")

    def extract_semantic_feat(self, tensor_x):
        """
        flatten the shape of feature tensor from super()
        """
        feat_tensor = super().extract_semantic_feat(tensor_x)
        feat = feat_tensor.reshape(feat_tensor.shape[0], -1)
        return feat

    def cal_logit_y(self, tensor_x):
        """
        calculate the logit for softmax classification
        """
        feat = self.extract_semantic_feat(tensor_x)
        logits = self.net_classifier(feat)
        return logits

    @store_args
    def __init__(self, list_str_y=None):
        """
        :param list_str_y: list of fixed order, each element is a class label
        """
        super().__init__()
        self.list_str_y = list_str_y
        self.perf_metric = None
        self.loss4gen_adv = nn.KLDivLoss(size_average=False)
        self._net_classifier = None

    def infer_y_vpicn(self, tensor):
        """
        :param tensor: input
        :return: vpicn
            v: vector of one-hot class label,
            p: vector of probability,
            i: class label index,
            c: confidence: maximum probability,
            n: list of name of class
        """
        with torch.no_grad():
            logit_y = self.cal_logit_y(tensor)
        vec_one_hot, prob, ind, confidence = logit2preds_vpic(logit_y)
        na_class = get_label_na(ind, self.list_str_y)
        return vec_one_hot, prob, ind, confidence, na_class

    @property
    def dim_y(self):
        """
        the class embedding dimension
        """
        return len(self.list_str_y)

    def cal_task_loss(self, tensor_x, tensor_y):
        """
        Calculate the task loss. Used within the `cal_loss` methods of models
        that are subclasses of `AModelClassif`. Cross entropy loss for
        classification is used here by default but could be modified by
        subclasses
        as necessary.

        :param tensor_x: input
        :param tensor_y: label
        :return: task loss
        """
        logit_y = self.cal_logit_y(tensor_x)
        if (tensor_y.shape[-1] == 1) | (len(tensor_y.shape) == 1):
            y_target = tensor_y
        else:
            _, y_target = tensor_y.max(dim=1)
        lc_y = F.cross_entropy(logit_y, y_target, reduction="none")
        # cross entropy always return a scalar, no need for inside instance reduction
        return lc_y

    def pred2file(self, loader_te, device, filename, metric_te, spliter="#"):
        """
        pred2file dump predicted label to file as sanity check
        """
        self.eval()
        model_local = self.to(device)
        logger = Logger.get_logger()
        for _, (x_s, y_s, *_, path4instance) in enumerate(loader_te):
            x_s, y_s = x_s.to(device), y_s.to(device)
            _, prob, *_ = model_local.infer_y_vpicn(x_s)
            list_pred_prob_list = prob.tolist()
            list_target_list = y_s.tolist()
            list_target_scalar = [
                np.asarray(label).argmax() for label in list_target_list
            ]
            tuple_zip = zip(path4instance, list_target_scalar, list_pred_prob_list)
            list_pair_path_pred = list(tuple_zip)
            with open(filename, "a", encoding="utf8") as handle_file:
                for list4one_obs_path_prob_target in list_pair_path_pred:
                    list_str_one_obs_path_target_predprob = [
                        str(ele) for ele in list4one_obs_path_prob_target
                    ]
                    str_line = (" " + spliter + " ").join(
                        list_str_one_obs_path_target_predprob
                    )
                    str_line = str_line.replace("[", "")
                    str_line = str_line.replace("]", "")
                    print(str_line, file=handle_file)
        logger.info(f"prediction saved in file {filename}")
        file_acc = self.read_prediction_file(filename, spliter)
        acc_metric_te = metric_te["acc"]
        flag1 = math.isclose(file_acc, acc_metric_te, rel_tol=1e-9, abs_tol=0.01)
        acc_raw1 = PerfClassif.cal_acc(self, loader_te, device)
        acc_raw2 = PerfClassif.cal_acc(self, loader_te, device)
        flag_raw_consistency = math.isclose(
            acc_raw1, acc_raw2, rel_tol=1e-9, abs_tol=0.01
        )
        flag2 = math.isclose(file_acc, acc_raw1, rel_tol=1e-9, abs_tol=0.01)
        if not (flag1 & flag2 & flag_raw_consistency):
            str_info = (
                f"inconsistent acc: \n"
                f"prediction file acc generated using the current model is {file_acc} \n"
                f"input torchmetric acc to the current function: {acc_metric_te} \n"
                f"raw acc 1 {acc_raw1} \n"
                f"raw acc 2 {acc_raw2} \n"
            )
            raise RuntimeError(str_info)
        return file_acc

    def read_prediction_file(self, filename, spliter):
        """
        check if the written fiel could calculate acc
        """
        with open(filename, "r", encoding="utf8") as handle_file:
            list_lines = [line.strip().split(spliter) for line in handle_file]
        count_correct = 0
        for line in list_lines:
            list_prob = [float(ele) for ele in line[2].split(",")]
            if np.array(list_prob).argmax() == int(line[1]):
                count_correct += 1
        acc = count_correct / len(list_lines)
        logger = Logger.get_logger()
        logger.info(f"accuracy from prediction file {acc}")
        return acc

    def cal_loss_gen_adv(self, x_natural, x_adv, vec_y):
        """
        calculate loss function for generation of adversarial images
        """
        x_adv.requires_grad_()
        with torch.enable_grad():
            logits_adv = self.cal_logit_y(x_adv)
            logits_natural = self.cal_logit_y(x_natural)
            prob_adv = F.log_softmax(logits_adv, dim=1)
            prob_natural = F.softmax(logits_natural, dim=1)
            loss_adv_gen_task = self.cal_task_loss(x_adv, vec_y)
            loss_adv_gen = self.loss4gen_adv(prob_adv, prob_natural)
        return loss_adv_gen + loss_adv_gen_task.sum()

    def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        for ERM to adapt to the interface of other regularized learners
        """
        device = tensor_x.device
        bsize = tensor_x.shape[0]
        return [torch.zeros(bsize).to(device)], [0.0]



"""
update hyper-parameters during training
"""
import numpy as np


class HyperSchedulerWarmupLinear:
    """
    HyperSchedulerWarmupLinear
    """

    def __init__(self, trainer, **kwargs):
        """
        kwargs is a dictionary with key the hyper-parameter name and its value
        """
        self.trainer = trainer
        self.dict_par_setpoint = kwargs
        self.total_steps = None

    def set_steps(self, total_steps):
        """
        set number of total_steps to gradually change optimization parameter
        """
        self.total_steps = total_steps

    def warmup(self, par_setpoint, epoch):
        """warmup.
        start from a small value of par to ramp up the steady state value using
        # total_steps
        :param epoch:
        """
        ratio = ((epoch + 1) * 1.0) / self.total_steps
        list_par = [par_setpoint, par_setpoint * ratio]
        par = min(list_par)
        return par

    def __call__(self, epoch):
        dict_rst = {}
        for key, val_setpoint in self.dict_par_setpoint.items():
            dict_rst[key] = self.warmup(val_setpoint, epoch)
        return dict_rst


class HyperSchedulerWarmupExponential(HyperSchedulerWarmupLinear):
    """
    HyperScheduler Exponential
    """

    def warmup(self, par_setpoint, epoch):
        """
        start from a small value of par to ramp up the steady state value using
        number of total_steps
        :param epoch:
        """
        percent_steps = ((epoch + 1) * 1.0) / self.total_steps
        denominator = 1.0 + np.exp(-10 * percent_steps)
        ratio = 2.0 / denominator - 1
        # percent_steps is 0, denom is 2, 2/denom is 1, ratio is 0
        # percent_steps is 1, denom is 1+exp(-10), 2/denom is 2/(1+exp(-10))=2, ratio is 1
        # exp(-10)=4.5e-5 is approximately 0
        # slowly increase the regularization weight from 0 to 1*alpha as epochs goes on
        parval = float(ratio * par_setpoint)
        return parval



"""
operations that all kinds of models should have
"""

import abc

from torch import nn

from domainlab import g_list_model_penalized_reg_agg


class AModel(nn.Module, metaclass=abc.ABCMeta):
    """
    operations that all models (classification, segmentation, seq2seq)
    """

    def __init__(self):
        super().__init__()
        self._decoratee = None
        self.list_d_tr = None
        self.visitor = None
        self._net_invar_feat = None

    def extend(self, model):
        """
        extend the loss of the decoratee
        """
        self._decoratee = model
        self.reset_feature_extractor(model.net_invar_feat)

    @property
    def metric4msel(self):
        """
        metric for model selection
        """
        raise NotImplementedError

    @property
    def multiplier4task_loss(self):
        """
        the multiplier for task loss is default to 1 except for vae family models
        """
        return 1.0

    def cal_loss(self, tensor_x, tensor_y, tensor_d=None, others=None):
        """
        calculate the loss
        """
        list_loss, list_multiplier = self.cal_reg_loss(
            tensor_x, tensor_y, tensor_d, others
        )
        loss_reg = self.list_inner_product(list_loss, list_multiplier)
        loss_task_alone = self.cal_task_loss(tensor_x, tensor_y)
        loss_task = self.multiplier4task_loss * loss_task_alone
        return loss_task + loss_reg, list_loss, loss_task_alone

    def list_inner_product(self, list_loss, list_multiplier):
        """
        compute inner product between list of regularization loss and multiplier
        - the length of the list is the number of regularizers
        - for each element of the list: the first dimension of the tensor is mini-batch
        return value of list_inner_product should keep the minibatch structure, thus aggregation
        here only aggregate along the list
        """
        list_tuple = zip(list_loss, list_multiplier)
        list_penalized_reg = [mtuple[0] * mtuple[1] for mtuple in list_tuple]
        tensor_batch_penalized_loss = g_list_model_penalized_reg_agg(list_penalized_reg)
        # return value of list_inner_product should keep the minibatch structure, thus aggregation
        # here only aggregate along the list
        return tensor_batch_penalized_loss

    @abc.abstractmethod
    def cal_task_loss(self, tensor_x, tensor_y):
        """
        Calculate the task loss

        :param tensor_x: input
        :param tensor_y: label
        :return: task loss
        """

    @abc.abstractmethod
    def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        task independent regularization loss for domain generalization
        """

    def cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        task independent regularization loss for domain generalization
        """
        loss_reg, mu = self._extend_loss(tensor_x, tensor_y, tensor_d, others)
        loss_reg_, mu_ = self._cal_reg_loss(tensor_x, tensor_y, tensor_d, others)
        if loss_reg is not None:
            return loss_reg_ + loss_reg, mu_ + mu
        return loss_reg_, mu_

    def _extend_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        combine losses from two models
        """
        if self._decoratee is not None:
            return self._decoratee.cal_reg_loss(tensor_x, tensor_y, tensor_d, others)
        return None, None

    def forward(self, tensor_x, tensor_y, tensor_d, others=None):
        """forward.

        :param x:
        :param y:
        :param d:
        """
        return self.cal_loss(tensor_x, tensor_y, tensor_d, others)

    def extract_semantic_feat(self, tensor_x):
        """
        extract semantic feature (not domain feature), note that
        extract semantic feature is an action, it is more general than
        calling a static network(module)'s forward function since
        there are extra action like reshape the tensor
        """
        if self._decoratee is not None:
            return self._decoratee.extract_semantic_feat(tensor_x)
        feat = self._net_invar_feat(tensor_x)
        return feat

    @property
    def net_invar_feat(self):
        """
        if exist, return a neural network for extracting invariant features
        """
        return self._net_invar_feat

    def reset_feature_extractor(self, net):
        """
        for two models to share the same neural network, the feature extractor has to be reset
        for classification, both feature extractor and classifier has to be reset
        """
        # note if net is None, which means the decoratee does not have net_invar_feat (can be
        # because there is tensor reshape during forward pass, which can not be represented
        # by a static neural network, in this case, we simply set self._net_invar_feat to be
        # None
        self._net_invar_feat = net
        self.reset_aux_net()

    def reset_aux_net(self):
        """
        after feature extractor being reset, the input dim of other networks like domain
        classification will also change (for commandline usage only)
        """
        # by default doing nothing

    def save(self, suffix=None):
        """
        persist model to disk
        """
        if self.visitor is None:
            return
        self.visitor.save(self, suffix)
        return

    def load(self, suffix=None):
        """
        load model from disk
        """
        if self.visitor is None:
            return None
        return self.visitor.load(suffix)

    def set_saver(self, visitor):
        self.visitor = visitor

    def dset_decoration_args_algo(self, args, ddset):
        """
        decorate dataset to get extra entries in load item, for instance, jigen need permutation index
        this parent class function delegate decoration to its decoratee
        """
        if self._decoratee is not None:
            return self._decoratee.dset_decoration_args_algo(args, ddset)
        return ddset



from domainlab.algos.observers.a_observer import AObVisitor


class ObVisitorCleanUp(AObVisitor):
    """
    decorator of observer, instead of using if and else to decide clean up or not, we use decorator
    """

    def __init__(self, observer):
        super().__init__()
        self.observer = observer

    def after_all(self):
        self.observer.after_all()
        self.observer.clean_up()

    def accept(self, trainer):
        self.observer.accept(trainer)

    def update(self, epoch):
        return self.observer.update(epoch)

    def clean_up(self):
        self.observer.clean_up()

    @property
    def model_sel(self):
        return self.observer.model_sel



"""
Base Class for XYD VAE Classify
"""
from domainlab.models.a_model_classif import AModelClassif
from domainlab.models.interface_vae_xyd import InterfaceVAEXYD
from domainlab.utils.utils_class import store_args


class VAEXYDClassif(AModelClassif, InterfaceVAEXYD):
    """
    Base Class for DIVA and HDUVA
    """

    @store_args
    def __init__(self, chain_node_builder, zd_dim, zy_dim, zx_dim, list_str_y):
        """
        :param chain_node_builder: constructed object
        """
        super().__init__(list_str_y)
        self.init()

    def extract_semantic_feat(self, tensor_x):
        zy_q_loc = self.encoder.infer_zy_loc(tensor_x)
        return zy_q_loc

    @property
    def multiplier4task_loss(self):
        """
        the multiplier for task loss is default to 1.0 except for vae family models
        """
        return self.gamma_y

    def _init_components(self):
        super()._init_components()
        self.add_module(
            "net_classif_y",
            self.chain_node_builder.construct_classifier(self.zy_dim, self.dim_y),
        )
        # property setter only for other object, internally, one shoud use _net_classifier
        self._net_classifier = self.net_classif_y



"""
Model Selection should be decoupled from
"""
from domainlab.algos.msels.c_msel_tr_loss import MSelTrLoss
from domainlab.utils.logger import Logger


class MSelValPerf(MSelTrLoss):
    """
    1. Model selection using validation performance
    2. Visitor pattern to trainer
    """

    def __init__(self, max_es):
        super().__init__(max_es)  # construct self.tr_obs (observer)
        self.reset()

    def reset(self):
        super().reset()
        self._best_val_acc = 0.0
        self._sel_model_te_acc = 0.0
        self._best_te_metric = 0.0

    @property
    def sel_model_te_acc(self):
        return self._sel_model_te_acc

    @property
    def best_val_acc(self):
        """
        decoratee best val acc
        """
        return self._best_val_acc

    @property
    def best_te_metric(self):
        """
        decoratee best test metric
        """
        return self._best_te_metric

    def update(self, clear_counter=False):
        """
        if the best model should be updated
        """
        flag = True
        if self.tr_obs.metric_val is None:
            return super().update(clear_counter)
        metric = self.tr_obs.metric_val[self.tr_obs.str_metric4msel]
        if self.tr_obs.metric_te is not None:
            metric_te_current = self.tr_obs.metric_te[self.tr_obs.str_metric4msel]
            self._best_te_metric = max(self._best_te_metric, metric_te_current)

        if metric > self._best_val_acc:  # update hat{model}
            # different from loss, accuracy should be improved:
            # the bigger the better
            self._best_val_acc = metric
            self.es_c = 0  # restore counter
            if self.tr_obs.metric_te is not None:
                metric_te_current = self.tr_obs.metric_te[self.tr_obs.str_metric4msel]
                self._sel_model_te_acc = metric_te_current

        else:
            self.es_c += 1
            logger = Logger.get_logger()
            logger.info(f"early stop counter: {self.es_c}")
            logger.info(
                f"val acc:{self.tr_obs.metric_val['acc']}, "
                + f"best validation acc: {self.best_val_acc}, "
                + f"corresponding to test acc: \
                        {self.sel_model_te_acc} / {self.best_te_metric}"
            )
            flag = False  # do not update best model
            if clear_counter:
                logger.info("clearing counter")
                self.es_c = 0
        return flag



"""
When class names and numbers does not match across different domains
"""
from torchvision import transforms

from domainlab.dsets.dset_subfolder import DsetSubFolder
from domainlab.dsets.utils_data import (
    DsetInMemDecorator,
    fun_img_path_loader_default,
    mk_fun_label2onehot,
)
from domainlab.tasks.b_task_classif import NodeTaskDictClassif
from domainlab.tasks.utils_task import DsetClassVecDecoratorImgPath
from domainlab.utils.logger import Logger


class NodeTaskFolder(NodeTaskDictClassif):
    """
    create dataset by loading files from an organized folder
    then each domain correspond to one dataset
    """

    @property
    def dict_domain2imgroot(self):
        """
        {"domain name":"xx/yy/zz"}
        """
        return self._dict_domains2imgroot

    @dict_domain2imgroot.setter
    def dict_domain2imgroot(self, dict_root):
        """
        {"domain name":"xx/yy/zz"}
        """
        if not isinstance(dict_root, dict):
            raise RuntimeError("input is not diciontary")
        self._dict_domains2imgroot = dict_root

    @property
    def extensions(self):
        """
        return allowed extensions
        """
        return self.dict_att["img_extensions"]

    @extensions.setter
    def extensions(self, str_format):
        self.dict_att["img_extensions"] = str_format

    def get_dset_by_domain(self, args, na_domain, split=False):
        if float(args.split):
            raise RuntimeError(
                "this task does not support spliting training domain yet"
            )
        if self._dict_domain_img_trans:
            trans = self._dict_domain_img_trans[na_domain]
            if na_domain not in self.list_domain_tr:
                trans = self.img_trans_te
        else:
            trans = transforms.ToTensor()
        dset = DsetSubFolder(
            root=self.dict_domain2imgroot[na_domain],
            list_class_dir=self.list_str_y,
            loader=fun_img_path_loader_default,
            extensions=self.extensions,
            transform=trans,
            target_transform=mk_fun_label2onehot(len(self.list_str_y)),
        )
        return dset, dset  # @FIXME: validation by default set to be training set


class NodeTaskFolderClassNaMismatch(NodeTaskFolder):
    """
    when the folder names of the same class from different domains have
    different names
    """

    def get_dset_by_domain(self, args, na_domain, split=False):
        if float(args.split):
            raise RuntimeError(
                "this task does not support spliting training domain yet"
            )
        logger = Logger.get_logger()
        logger.info(f"reading domain: {na_domain}")
        domain_class_dirs = self._dict_domain_folder_name2class[na_domain].keys()
        if self._dict_domain_img_trans:
            trans = self._dict_domain_img_trans[na_domain]
            if na_domain not in self.list_domain_tr:
                trans = self.img_trans_te
        else:
            trans = transforms.ToTensor()

        ext = None if self.extensions is None else self.extensions[na_domain]
        dset = DsetSubFolder(
            root=self.dict_domain2imgroot[na_domain],
            list_class_dir=list(domain_class_dirs),
            loader=fun_img_path_loader_default,
            extensions=ext,
            transform=trans,
            target_transform=mk_fun_label2onehot(len(self.list_str_y)),
        )
        # dset.path2imgs
        dict_folder_name2class_global = self._dict_domain_folder_name2class[na_domain]
        dset = DsetClassVecDecoratorImgPath(
            dset, dict_folder_name2class_global, self.list_str_y
        )
        # Always use the DsetInMemDecorator at the last step
        # since it does not have other needed attributes in bewteen
        if args.dmem:
            dset = DsetInMemDecorator(dset, na_domain)
        return dset, dset  # @FIXME: validation by default set to be training set



"""
Random seed should be set from command line to ensure reproducibility:
https://pytorch.org/docs/stable/notes/randomness.html
https://discuss.pytorch.org/t/difference-between-torch-manual-seed-and-torch-cuda-manual-seed/13848/6
"""
import numpy as np
import torch


def set_seed(seed):
    torch.manual_seed(int(seed))
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # for reproducibility, this has to be False: benchmark mode is good
    # whenever your input sizes for your network do not vary. This way,
    # cudnn will look for the optimal set of algorithms for that particular
    # configuration (which takes some time). This usually leads to faster
    # runtime.
    np.random.seed(int(seed))



"""
In PyTorch, images are represented as [channels, height, width]
"""
import torch
import torch.nn as nn

from domainlab.compos.nn_zoo.nn import DenseNet
from domainlab.compos.utils_conv_get_flat_dim import get_flat_dim


def mk_conv_bn_relu_pool(i_channel, conv_stride=1, max_pool_stride=2):
    """
    Convolution, batch norm, maxpool_2d
    Convolution with maxpool_2d as last operation
    :param i_channel:
    :param conv_stride:
    :param max_pool_stride:
    """
    conv_net = nn.Sequential(
        nn.Conv2d(i_channel, 32, kernel_size=5, stride=conv_stride, bias=False),
        nn.BatchNorm2d(32),
        nn.ReLU(),
        nn.MaxPool2d(2, stride=max_pool_stride),
        nn.Conv2d(32, 64, kernel_size=5, stride=conv_stride, bias=False),
        nn.BatchNorm2d(64),
        nn.ReLU(),
        nn.MaxPool2d(2, stride=max_pool_stride),
    )
    torch.nn.init.xavier_uniform_(conv_net[0].weight)
    torch.nn.init.xavier_uniform_(conv_net[4].weight)
    return conv_net


class NetConvBnReluPool2L(nn.Module):
    def __init__(self, isize, conv_stride, dim_out_h):
        """
        :param dim_out_h:
        """
        super().__init__()
        i_c, i_h, i_w = isize
        self.conv_net = mk_conv_bn_relu_pool(i_c, conv_stride)
        self.hdim = get_flat_dim(self.conv_net, i_c, i_h, i_w)
        self.layer_last = nn.Linear(self.hdim, dim_out_h)

    def forward(self, tensor_x):
        """
        :param tensor_x: image
        """
        conv_out = self.conv_net(
            tensor_x
        )  # conv-bn-relu-pool-conv-bn-relu-pool(no activation)
        flat = conv_out.view(-1, self.hdim)  # 1024 =   64 * (4*4)
        hidden = self.layer_last(flat)
        return hidden


class NetConvDense(nn.Module):
    """
    - For direct topic inference
    - For custom erm, which is extracting the path of VAE from encoder
      until classifier. note in encoder, there is extra layer of hidden to mean
      and scale, in this component, it is replaced with another hidden layer.
    """

    def __init__(self, isize, conv_stride, dim_out_h, args, dense_layer=None):
        """
        :param dim_out_h:
        """
        super().__init__()
        i_c, i_h, i_w = isize
        self.conv_net = mk_conv_bn_relu_pool(i_c, conv_stride)
        ###
        self.hdim = get_flat_dim(self.conv_net, i_c, i_h, i_w)
        if dense_layer is None:
            self.dense_layers = DenseNet(self.hdim, out_hidden_size=dim_out_h)
        else:
            self.dense_layers = dense_layer

    def forward(self, tensor_x):
        """
        :param tensor_x: image
        """
        conv_out = self.conv_net(
            tensor_x
        )  # conv-bn-relu-pool-conv-bn-relu-pool(no activation)
        flat = conv_out.view(-1, self.hdim)  # 1024 =   64 * (4*4)
        hidden = self.dense_layers(flat)
        return hidden



"this script can be used to download the pacs dataset"
import os
import tarfile
from zipfile import ZipFile

import gdown


def stage_path(data_dir, name):
    """
    creates the path to data_dir/name
    if it does not exist already
    """
    full_path = os.path.join(data_dir, name)

    if not os.path.exists(full_path):
        os.makedirs(full_path)

    return full_path


def download_and_extract(url, dst, remove=True):
    """
    downloads and extracts the data behind the url
    and saves it at dst
    """
    gdown.download(url, dst, quiet=False)

    if dst.endswith(".tar.gz"):
        with open(dst, "r:gz") as tar:
            tar.extractall(os.path.dirname(dst))
        tar.close()

    if dst.endswith(".tar"):
        with open(dst, "r:") as tar:
            tar.extractall(os.path.dirname(dst))
        tar.close()

    if dst.endswith(".zip"):
        zfile = ZipFile(dst, "r")
        zfile.extractall(os.path.dirname(dst))
        zfile.close()

    if remove:
        os.remove(dst)


def download_pacs(data_dir):
    """
    download and extract dataset pacs.
    Dataset is saved at location data_dir
    """
    full_path = stage_path(data_dir, "PACS")

    download_and_extract(
        "https://drive.google.com/uc?id=1JFr8f805nMUelQWWmfnJR3y4_SYoN5Pd",
        os.path.join(data_dir, "PACS.zip"),
    )

    os.rename(os.path.join(data_dir, "kfold"), full_path)


if __name__ == "__main__":
    download_pacs("../pacs")



"""
Hierarchical Domain Unsupervised Variational Auto-Encoding
"""
import torch
from torch.distributions import Dirichlet

from domainlab import g_inst_component_loss_agg
from domainlab.models.model_vae_xyd_classif import VAEXYDClassif
from domainlab.utils.utils_class import store_args


def mk_hduva(parent_class=VAEXYDClassif):
    """
    Instantiate a Hierarchical Domain Unsupervised VAE (HDUVA) with arbitrary task loss.

    Details:
        The created model builds on a generative approach within the framework of variational
        autoencoders to facilitate generalization to new domains without supervision. HDUVA learns
        representations that disentangle domain-specific information from class-label specific
        information even in complex settings where domain structure is not observed during training.
        Here for, latent variables are introduced, representing the information for the classes,
        domains and the residual variance of the inputs, respectively. The domain structure is
        modelled by a hierarchical level and another latent variable, denoted as topic.
        Two encoder networks are trained. One for converting an image to be compatible with the
        latent spaces of the domains and another one for converting an image to a topic
        distribution. The overall objective is constructed by adding an additional weighted term to
        the ELBO loss. One benefit of this model is that the domain information during training can
        be incomplete.
        For more details, see: Sun, Xudong, and Buettner, Florian.
        "Hierarchical variational auto-encoding for unsupervised domain generalization."
        arXiv preprint arXiv:2101.09436 (2021).

    Args:
        parent_class: Class object determining the task type. Defaults to VAEXYDClassif.

    Returns:
        ModelHDUVA: model inheriting from parent class.

    Input Parameters:
        zd_dim: size of latent space for domain-specific information (int),
        zy_dim: size of latent space for class-specific information (int),
        zx_dim: size of latent space for residual variance (int, defaults to 0),
        chain_node_builder: an object which can build different maps via neural network,
        list_str_y: list of labels (list of strings),
        gamma_d: weighting term for domain classificaiton loss
        gamma_y: weighting term for additional term in ELBO loss (float),
        beta_d: weighting term for the domain component of ELBO loss (float),
        beta_x: weighting term for residual variation component of ELBO loss (float),
        beta_y: weighting term for class component of ELBO loss (float),
        beta_t: weighting term for the topic component of ELBO loss (float),
        device: device to which the model should be moved (cpu or gpu),
        topic_dim: size of latent space for topics (int, defaults to 3)
    """

    class ModelHDUVA(parent_class):
        """
        Hierarchical Domain Unsupervised Variational Auto-Encoding
        """

        def hyper_update(self, epoch, fun_scheduler):
            """hyper_update.

            :param epoch:
            :param fun_scheduler:
            """
            dict_rst = fun_scheduler(
                epoch
            )  # the __call__ function of hyper-para-scheduler object
            self.beta_d = dict_rst["beta_d"]
            self.beta_y = dict_rst["beta_y"]
            self.beta_x = dict_rst["beta_x"]
            self.beta_t = dict_rst["beta_t"]

        def hyper_init(self, functor_scheduler):
            """hyper_init.
            :param functor_scheduler:
            """
            # calling the constructor of the hyper-parameter-scheduler class, so that this scheduler
            # class build a dictionary {"beta_d":self.beta_d, "beta_y":self.beta_y}
            # constructor signature is def __init__(self, **kwargs):
            return functor_scheduler(
                trainer=None,
                beta_d=self.beta_d,
                beta_y=self.beta_y,
                beta_x=self.beta_x,
                beta_t=self.beta_t,
            )

        @store_args
        def __init__(
            self,
            chain_node_builder,
            zy_dim,
            zd_dim,
            list_str_y,
            gamma_d,
            gamma_y,
            beta_d,
            beta_x,
            beta_y,
            beta_t,
            device,
            zx_dim=0,
            topic_dim=3,
            multiplier_recon=1.0,
        ):
            """ """
            super().__init__(chain_node_builder, zd_dim, zy_dim, zx_dim, list_str_y)

            # topic to zd follows Gaussian distribution
            self.add_module(
                "net_p_zd",
                self.chain_node_builder.construct_cond_prior(
                    self.topic_dim, self.zd_dim
                ),
            )

        # override interface
        def _init_components(self):
            """
            q(z|x)
            p(zy)
            q_{classif}(zy)
            """
            self.add_module(
                "encoder",
                self.chain_node_builder.build_encoder(self.device, self.topic_dim),
            )
            self.add_module(
                "decoder", self.chain_node_builder.build_decoder(self.topic_dim)
            )
            self.add_module(
                "net_p_zy",
                self.chain_node_builder.construct_cond_prior(self.dim_y, self.zy_dim),
            )
            self.add_module(
                "net_classif_y",
                self.chain_node_builder.construct_classifier(self.zy_dim, self.dim_y),
            )
            self._net_classifier = self.net_classif_y

        def init_p_topic_batch(self, batch_size, device):
            """
            flat prior
            """
            prior = Dirichlet(torch.ones(batch_size, self.topic_dim).to(device))
            return prior

        def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d=None, others=None):
            q_topic, topic_q, qzd, zd_q, qzx, zx_q, qzy, zy_q = self.encoder(tensor_x)

            batch_size = zd_q.shape[0]
            device = zd_q.device

            p_topic = self.init_p_topic_batch(batch_size, device)

            # @FIXME: does monte-carlo KL makes the performance unstable?
            # from torch.distributions import kl_divergence

            # zy KL divergence

            if (tensor_y.shape[-1] == 1) | (len(tensor_y.shape) == 1):
                tensor_y_onehot = torch.nn.functional.one_hot(
                    tensor_y, num_classes=len(self.list_str_y)
                )
                tensor_y_onehot = tensor_y_onehot.to(torch.float32)
            else:
                tensor_y_onehot = tensor_y

            p_zy = self.net_p_zy(tensor_y_onehot)
            zy_p_minus_zy_q = g_inst_component_loss_agg(
                p_zy.log_prob(zy_q) - qzy.log_prob(zy_q), 1
            )

            # zx KL divergence
            zx_p_minus_q = torch.zeros_like(zy_p_minus_zy_q)
            if self.zx_dim > 0:
                p_zx = self.init_p_zx4batch(batch_size, device)
                zx_p_minus_q = g_inst_component_loss_agg(
                    p_zx.log_prob(zx_q) - qzx.log_prob(zx_q), 1
                )

            # zd KL diverence
            p_zd = self.net_p_zd(topic_q)
            zd_p_minus_q = g_inst_component_loss_agg(
                p_zd.log_prob(zd_q) - qzd.log_prob(zd_q), 1
            )

            # topic KL divergence
            # @FIXME: why topic is still there?
            topic_p_minus_q = p_topic.log_prob(topic_q) - q_topic.log_prob(topic_q)

            # reconstruction
            z_concat = self.decoder.concat_ytdx(zy_q, topic_q, zd_q, zx_q)
            loss_recon_x, _, _ = self.decoder(z_concat, tensor_x)
            return [
                loss_recon_x,
                zx_p_minus_q,
                zy_p_minus_zy_q,
                zd_p_minus_q,
                topic_p_minus_q,
            ], [
                self.multiplier_recon,
                -self.beta_x,
                -self.beta_y,
                -self.beta_d,
                -self.beta_t,
            ]

        def extract_semantic_feat(self, tensor_x):
            """
            :param tensor_x:
            """
            zy_q_loc = self.encoder.infer_zy_loc(tensor_x)
            return zy_q_loc

    return ModelHDUVA



import torch.nn as nn

from domainlab.compos.vae.compos.encoder_dirichlet import EncoderH2Dirichlet
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter


class EncoderImg2TopicDistri(nn.Module):
    """
    image to topic distribution  (not image to topic hidden representation
    used by another path)
    """

    def __init__(self, isize, num_topics, device, args):
        """__init__.

        :param isize:
        :param num_topics:
        :param device:
        """
        super().__init__()
        self.device = device

        # image->h_topic->(batchnorm, exp)[alpha,topic]

        net_builder = FeatExtractNNBuilderChainNodeGetter(
            args=args,
            arg_name_of_net="nname_encoder_x2topic_h",
            arg_path_of_net="npath_encoder_x2topic_h",
        )()

        self.add_module(
            "layer_img2hidden",
            net_builder.init_business(
                flag_pretrain=True,
                isize=isize,
                remove_last_layer=False,
                dim_out=num_topics,
                args=args,
            ),
        )

        # h_image->[alpha,topic]
        self.add_module(
            "layer_hidden2dirichlet",
            EncoderH2Dirichlet(dim_topic=num_topics, device=self.device),
        )

    def forward(self, x):
        """forward.

        :param x:
        """
        # image->h_image
        h_img_dir = self.layer_img2hidden(x)
        # h_image->alpha
        q_topic, topic_q = self.layer_hidden2dirichlet(h_img_dir)
        return q_topic, topic_q



"""
Base Class for trainer
"""
import abc

from torch import optim

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler


def mk_opt(model, aconf):
    """
    create optimizer
    """
    if model._decoratee is None:
        optimizer = optim.Adam(model.parameters(), lr=aconf.lr)
    else:
        var1 = model.parameters()
        var2 = model._decoratee.parameters()
        set_param = set(list(var1) + list(var2))
        list_par = list(set_param)
        # optimizer = optim.Adam([var1, var2], lr=aconf.lr)
        # optimizer = optim.Adam([
        #    {'params': model.parameters()},
        #    {'params': model._decoratee.parameters()}
        # ], lr=aconf.lr)
        optimizer = optim.Adam(list_par, lr=aconf.lr)
    return optimizer


class AbstractTrainer(AbstractChainNodeHandler, metaclass=abc.ABCMeta):
    """
    Algorithm director that controls the data flow
    """

    @property
    def p_na_prefix(self):
        """
        common prefix for Trainers
        """
        return "Trainer"

    def extend(self, trainer):
        """
        extend current trainer with another trainer
        """
        self._decoratee = trainer

    def __init__(self, successor_node=None, extend=None):
        """__init__.
        :param successor_node:
        """
        super().__init__(successor_node)
        self._model = None
        self._decoratee = extend
        self.task = None
        self.observer = None
        self.device = None
        self.aconf = None
        #
        self.loader_tr = None
        self.loader_te = None
        self.num_batches = None
        self.flag_update_hyper_per_epoch = None
        self.flag_update_hyper_per_batch = None
        self.epo_loss_tr = None
        self.epo_reg_loss_tr = None
        self.epo_task_loss_tr = None
        self.counter_batch = None
        self.hyper_scheduler = None
        self.optimizer = None
        self.exp = None
        # matchdg
        self.lambda_ctr = None
        self.flag_stop = None
        self.flag_erm = None
        self.tensor_ref_domain2each_domain_x = None
        self.tensor_ref_domain2each_domain_y = None
        self.base_domain_size = None
        self.tuple_tensor_ref_domain2each_y = None
        self.tuple_tensor_refdomain2each = None
        # mldg
        self.inner_trainer = None
        self.loader_tr_source_target = None
        self.flag_initialized = False

    @property
    def model(self):
        """
        property model, which can be another trainer or model
        """
        return self.get_model()

    @model.setter
    def model(self, model):
        self._model = model

    @property
    def str_metric4msel(self):
        """
        metric for model selection
        """
        return self.model.metric4msel

    @property
    def list_tr_domain_size(self):
        """
        get a list of training domain size
        """
        train_domains = self.task.list_domain_tr
        return [len(self.task.dict_dset_tr[key]) for key in train_domains]

    @property
    def decoratee(self):
        if self._decoratee is None:
            return self.model
        return self._decoratee

    def init_business(self, model, task, observer, device, aconf, flag_accept=True):
        """
        model, task, observer, device, aconf
        """
        # Note self.decoratee can be both model and trainer,
        # but self._decoratee can only be trainer!
        if self._decoratee is not None:
            self._decoratee.init_business(
                model, task, observer, device, aconf, flag_accept
            )
        self.model = model
        self.task = task
        self.task.init_business(trainer=self, args=aconf)
        self.model.list_d_tr = self.task.list_domain_tr
        self.observer = observer
        self.device = device
        self.aconf = aconf
        #
        self.loader_tr = task.loader_tr
        self.loader_te = task.loader_te

        if flag_accept:
            self.observer.accept(self)

        self.model = self.model.to(device)
        #
        self.num_batches = len(self.loader_tr)
        self.flag_update_hyper_per_epoch = False
        self.flag_update_hyper_per_batch = False
        self.epo_loss_tr = None
        self.hyper_scheduler = None
        self.reset()
        self.flag_initialized = True

    def reset(self):
        """
        make a new optimizer to clear internal state
        """
        self.optimizer = mk_opt(self.model, self.aconf)

    @abc.abstractmethod
    def tr_epoch(self, epoch):
        """
        :param epoch:
        """

    def before_batch(self, epoch, ind_batch):
        """
        :param epoch:
        :param ind_batch:
        """
        return

    def after_batch(self, epoch, ind_batch):
        """
        :param epoch:
        :param ind_batch:
        """
        return

    @abc.abstractmethod
    def before_tr(self):
        """
        before training, probe model performance
        """

    def post_tr(self):
        """
        after training
        """
        self.observer.after_all()

    @property
    def name(self):
        """
        get the name of the algorithm
        """
        na_prefix = self.p_na_prefix
        len_prefix = len(na_prefix)
        na_class = type(self).__name__
        if na_class[:len_prefix] != na_prefix:
            raise RuntimeError(
                "Trainer builder node class must start with ",
                na_prefix,
                "the current class is named: ",
                na_class,
            )
        return type(self).__name__[len_prefix:].lower()

    def is_myjob(self, request):
        """
        :param request: string
        """
        return request == self.name

    def get_model(self):
        """
        recursively get the "real" model from trainer
        """
        if "trainer" not in str(type(self._model)).lower():
            return self._model
        return self._model.get_model()

    def cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        decorate trainer regularization loss
        combine losses of current trainer with self._model.cal_reg_loss, which
        can be either a trainer or a model
        """
        list_reg_model, list_mu_model = self.decoratee.cal_reg_loss(
            tensor_x, tensor_y, tensor_d, others
        )
        assert len(list_reg_model) == len(list_mu_model)

        list_reg_trainer, list_mu_trainer = self._cal_reg_loss(
            tensor_x, tensor_y, tensor_d, others
        )
        assert len(list_reg_trainer) == len(list_mu_trainer)

        list_loss = list_reg_model + list_reg_trainer
        list_mu = list_mu_model + list_mu_trainer
        return list_loss, list_mu

    def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        interface for each trainer to implement
        """
        return [], []

    def dset_decoration_args_algo(self, args, ddset):
        """
        decorate dataset to get extra entries in load item, for instance,
        jigen need permutation index
        this parent class function delegate decoration to its decoratee
        """
        if self._decoratee is not None:
            return self._decoratee.dset_decoration_args_algo(args, ddset)
        return ddset



from domainlab.compos.a_nn_builder import AbstractFeatExtractNNBuilderChainNode
from domainlab.compos.nn_zoo.net_conv_conv_bn_pool_2 import NetConvBnReluPool2L


def mkNodeFeatExtractNNBuilderNameConvBnRelu2(arg_name4net, arg_val, conv_stride):
    """mkNodeFeatExtractNNBuilderNameConvBnRelu2.
    In chain of responsibility selection of neural network, reuse code to add
    more possibilities of neural network of the same family.
    :param arg_name4net: name of nn in args
    :param arg_val: the registered name of the neural network to be added
    :param conv_stride: should be 1 for 28*28 images
    :param i_c:
    :param i_h:
    :param i_w:
    """

    class _NodeFeatExtractNNBuilderConvBnRelu2L(AbstractFeatExtractNNBuilderChainNode):
        """NodeFeatExtractNNBuilderConvBnRelu2L."""

        def init_business(
            self, dim_out, args, isize, flag_pretrain=None, remove_last_layer=False
        ):
            """
            :param flag_pretrain
            """
            self.net_feat_extract = NetConvBnReluPool2L(
                isize=isize, conv_stride=conv_stride, dim_out_h=dim_out
            )
            return self.net_feat_extract

        def is_myjob(self, args):
            """is_myjob.
            :param args: command line arguments:
                "--nname": name of the torchvision model
            """
            arg_name = getattr(args, arg_name4net)
            return arg_name == arg_val

    return _NodeFeatExtractNNBuilderConvBnRelu2L



"""
This module contains 3 classes inheriting:
    ExpProtocolAggWriter(AggWriter(ExpModelPersistVisitor))
"""
import copy
import datetime
import os
from pathlib import Path

import numpy as np
import torch
from sklearn.metrics import ConfusionMatrixDisplay

from domainlab.utils.get_git_tag import get_git_tag
from domainlab.utils.logger import Logger


class ExpModelPersistVisitor:
    """
    This class couples with Task class attributes
    """

    model_dir = "saved_models"
    model_suffix = ".model"

    def __init__(self, host):
        """
        1. create new attributes like model names
        2. all dependencies in contructor
        3. do not change the sequence of the lines!
        since later lines depends on earlier definitions
        """
        self.host = host
        self.out = host.args.out
        self.model_dir = os.path.join(self.out, ExpModelPersistVisitor.model_dir)
        self.git_tag = get_git_tag()
        self.task_name = self.host.task.get_na(self.host.args.tr_d, self.host.args.te_d)
        self.algo_name = self.host.args.model
        self.seed = self.host.args.seed
        self.model_name = self.mk_model_na(self.git_tag)
        self.model_path = os.path.join(
            self.model_dir, self.model_name + ExpModelPersistVisitor.model_suffix
        )

        Path(os.path.dirname(self.model_path)).mkdir(parents=True, exist_ok=True)
        self.model = copy.deepcopy(self.host.trainer.model)
        # although deepcopy in contructor is expensive, but
        # execute copy.deepcopy(self.host.trainer.model) after training will cause thread lock
        # if self.host.trainer has tensorboard writer, see
        # https://github.com/marrlab/DomainLab/issues/673

    def mk_model_na(self, tag=None, dd_cut=19):
        """
        :param tag: for git commit hash for example
        """
        if tag is None:
            tag = "tag"
        suffix_t = str(datetime.datetime.now())[:dd_cut].replace(" ", "_")
        suffix_t = suffix_t.replace("-", "md_")
        suffix_t = suffix_t.replace(":", "_")
        list4mname = [
            self.task_name,
            self.algo_name,
            tag,
            suffix_t,
            "seed",
            str(self.seed),
        ]
        # the sequence of components (e.g. seed in the last place)
        # in model name is not crutial
        model_name = "_".join(list4mname)
        if self.host.args.debug:
            model_name = "debug_" + model_name
        slurm = os.environ.get("SLURM_JOB_ID")
        if slurm:
            model_name = model_name + "_" + slurm
        logger = Logger.get_logger()
        logger.info(f"model name: {model_name}")
        return model_name

    def save(self, model, suffix=None):
        """
        :param model:
        """
        file_na = self.model_path
        if suffix is not None:
            file_na = "_".join([file_na, suffix])
        torch.save(copy.deepcopy(model.state_dict()), file_na)
        # checkpoint = {'model': Net(), '
        # state_dict': model.state_dict(),
        # 'optimizer' :optimizer.state_dict()}
        # torch.save(checkpoint, 'Checkpoint.pth')

    def remove(self, suffix=None):
        """
        remove model after use
        """
        file_na = self.model_path
        if suffix is not None:
            file_na = "_".join([file_na, suffix])
        os.remove(file_na)

    def load(self, suffix=None):
        """
        load pre-defined model name from disk
        the save function is the same class so to ensure load will ways work
        """
        path = self.model_path
        if suffix is not None:
            path = "_".join([self.model_path, suffix])
        # due to tensorboard writer in trainer.scheduler,
        # copy.deepcopy(self.host.trainer.model) will cause thread lock
        # see https://github.com/marrlab/DomainLab/issues/673
        self.model.load_state_dict(torch.load(path, map_location="cpu"))
        # without separate self.model and self.model_suffixed,
        # it will cause accuracy inconsistent problems since the content of self.model
        # can be overwritten when the current function is called another time and self.model
        # is not deepcopied
        # However, deepcopy is also problematic when it is executed too many times
        return copy.deepcopy(self.model)
        # instead of deepcopy, one could also have multiple copies of model in constructor, but this
        # does not adhere the lazy principle.

    def clean_up(self):
        self.host.clean_up()


class AggWriter(ExpModelPersistVisitor):
    """
    1. aggregate results to text file.
    2. all dependencies are in the constructor
    """

    def __init__(self, host):
        super().__init__(host)
        self.agg_tag = self.host.args.aggtag
        self.exp_tag = self.host.args.exptag
        self.debug = self.host.args.debug
        self.has_first_line = False
        self.list_cols = None

    def first_line(self, dict_cols):
        """
        generate header of the results aggregation file
        """
        self.list_cols = list(dict_cols.keys())
        # @FIXME: will be list be the same order each time?
        str_line = ", ".join(self.list_cols)
        if not os.path.isfile(self.get_fpath()):
            self.to_file(str_line)
        self.has_first_line = True

    def __call__(self, dict_metric):
        line, confmat, confmat_filename = self._gen_line(dict_metric)
        self.to_file(line)
        if not self.host.args.no_dump:
            self.confmat_to_file(confmat, confmat_filename)

    def get_cols(self):
        """
        call the same function to always get the same columns configuration
        """
        epos_name = "epos"
        dict_cols = {
            "algo": self.algo_name,
            epos_name: None,
            "seed": self.seed,
            "aggtag": self.agg_tag,
            # algorithm configuration for instance
            "mname": "mname_" + self.model_name,
            "commit": "commit_" + self.git_tag,
        }
        return dict_cols, epos_name

    def _gen_line(self, dict_metric):
        dict_cols, epos_name = self.get_cols()
        dict_cols.update(dict_metric)
        confmat = dict_cols.pop("confmat")
        confmat_filename = dict_cols.get("mname", None)  # return None if not found
        # @FIXME: strong dependency on host attribute name
        dict_cols.update({epos_name: self.host.epoch_counter})
        if not self.has_first_line:
            self.first_line(dict_cols)
        list_str = [str(dict_cols[key]) for key in self.list_cols]
        str_line = ", ".join(list_str)
        return str_line, confmat, confmat_filename

    def get_fpath(self, dirname="aggrsts"):
        """
        for writing and reading, the same function is called to ensure name
        change in the future will not break the software
        """
        list4fname = [
            self.task_name,
            self.exp_tag,
        ]
        fname = "_".join(list4fname) + ".csv"
        if self.debug:
            fname = "_".join(["debug_agg", fname])
        file_path = os.path.join(self.out, dirname, fname)
        return file_path

    def to_file(self, str_line):
        """
        :param str_line:
        """
        file_path = self.get_fpath()
        Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)
        logger = Logger.get_logger()
        logger.info(f"results aggregation path: {file_path}")
        with open(file_path, "a", encoding="utf8") as f_h:
            print(str_line, file=f_h)

    def confmat_to_file(self, confmat, confmat_filename):
        """Save confusion matrix as a figure

        Args:
            confmat: confusion matrix.
        """
        disp = ConfusionMatrixDisplay(confmat)
        disp = disp.plot(cmap="gray")
        file_path = self.get_fpath()
        # @FIXME: although removesuffix is safe when suffix does not exist,
        # we would like to have ".csv" live in some configuraiton file in the future.
        file_path = file_path.removesuffix(".csv")
        # if prefix does not exist, string remain unchanged.
        # @FIXME: still we want to have mname_ as a variable defined in some
        # configuration file in the future.
        confmat_filename = confmat_filename.removeprefix("mname_")
        file_path = os.path.join(
            os.path.dirname(file_path), f"{confmat_filename}_conf_mat.png"
        )
        logger = Logger.get_logger()
        logger.info(f"confusion matrix saved in file: {file_path}")
        disp.figure_.savefig(file_path)


class ExpProtocolAggWriter(AggWriter):
    """
    AggWriter tailored to experimental protocol
    Output contains additionally index, exp task, te_d and params.
    """

    def get_cols(self):
        """columns"""
        epos_name = "epos"
        dict_cols = {
            "param_index": self.host.args.param_index,
            "method": self.host.args.benchmark_task_name,
            "mname": "mname_" + self.model_name,
            "commit": "commit_" + self.git_tag,
            "algo": self.algo_name,
            epos_name: None,
            "te_d": self.host.args.te_d,
            "seed": self.seed,
            "params": f'"{self.host.args.params}"',
        }
        return dict_cols, epos_name

    def get_fpath(self, dirname=None):
        """filepath"""
        if dirname is None:
            return self.host.args.result_file
        return super().get_fpath(dirname)

    def confmat_to_file(self, confmat, confmat_filename):
        """Save confusion matrix as a figure

        Args:
            confmat: confusion matrix.
        """
        path4file = self.get_fpath()
        index = os.path.basename(path4file)
        path4file = os.path.dirname(os.path.dirname(path4file))
        # if prefix does not exist, string remain unchanged.
        confmat_filename = confmat_filename.removeprefix("mname_")
        path4file = os.path.join(path4file, "confusion_matrix")
        os.makedirs(path4file, exist_ok=True)
        file_path = os.path.join(path4file, f"{index}.txt")
        with open(file_path, "a", encoding="utf8") as f_h:
            print(confmat_filename, file=f_h)
            for line in np.matrix(confmat):
                np.savetxt(f_h, line, fmt="%.2f")
        logger = Logger.get_logger()
        logger.info(f"confusion matrix saved in file: {file_path}")



"""
generate the benchmark plots by calling the gen_bencmark_plots(...) function
"""
import os
from ast import literal_eval  # literal_eval can safe evaluate python expression

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from domainlab.utils.logger import Logger

matplotlib.use("Agg")

# header of the csv file:
# param_index, task, algo, epos, te_d, seed, params, acc, precision, recall, specificity, f1, auroc

COLNAME_METHOD = "method"
COLNAME_IDX_PARAM = "param_index"
COLNAME_PARAM = "params"
G_DF_TASK_COL = 1  # column in which the method name is saved
G_DF_PLOT_COL_METRIC_START = 9  # first 0-6 columns are not metric


def gen_benchmark_plots(
    agg_results: str, output_dir: str, use_param_index: bool = True
):
    """
    generate the benchmark plots from a csv file containing the aggregated restults.
    The csv file must have the columns:
    [param_index, task, algo, epos, te_d, seed, params, ...]
    all columns after seed are intrepreted as objectives of the results, they can e.g.
    be acc, precision, recall, specificity, f1, auroc.

    agg_results: path to the csv file
    output_dir: path to a folder which shall contain the results
    skip_gen: Skips the actual plotting, used to speed up testing.
    """
    raw_df = pd.read_csv(
        agg_results,
        index_col=False,
        converters={COLNAME_PARAM: literal_eval},
        # literal_eval can safe evaluate python expression
        skipinitialspace=True,
    )

    raw_df[COLNAME_PARAM] = round_vals_in_dict(
        raw_df[[COLNAME_IDX_PARAM, COLNAME_PARAM]], use_param_index
    )

    # generating plot
    gen_plots(raw_df, output_dir, use_param_index)


def round_vals_in_dict(df_column_in, use_param_index):
    """
    replaces the dictionary by a string containing only the significant digits of the hyperparams
    or (if use_param_index = True) by the parameter index
    df_column_in: columns of the dataframe containing the param index and the dictionary of
        hyperparams in the form [param_index, params]
    use_param_index: usage of param_index instead of exact values
    """
    df_column = df_column_in.copy()
    df_column_out = df_column_in[COLNAME_IDX_PARAM].copy()
    df_column_out = df_column_out.astype(str)
    for i in range(df_column.shape[0]):
        if not use_param_index:
            string = ""
            for num, val in enumerate(list(df_column[COLNAME_PARAM][i].values())):
                key = list(df_column[COLNAME_PARAM][i].keys())[num]
                val = np.format_float_scientific(
                    val, precision=1, unique=False, trim="0"
                )
                string += str(key) + ": " + str(val) + ", "
            df_column_out[i] = string[:-2]  # remove ', ' from the end of the string
        else:
            string = "idx: " + str(df_column[COLNAME_IDX_PARAM][i])
            df_column_out[i] = string
    return df_column_out


def gen_plots(dataframe: pd.DataFrame, output_dir: str, use_param_index: bool):
    """
    dataframe: dataframe with columns
    ['param_index','task',' algo',' epos',' te_d',' seed',' params',' acc','precision',...]
    """
    os.makedirs(output_dir, exist_ok=True)
    obj = dataframe.columns[G_DF_PLOT_COL_METRIC_START:]

    # boxplots
    for objective in obj:
        boxplot(
            dataframe, objective, file=output_dir + "/variational_plots/" + objective
        )

    # scatterplot matrices
    scatterplot_matrix(
        dataframe,
        use_param_index,
        file=output_dir + "/sp_matrix_reg.png",
        kind="reg",
        distinguish_param_setups=False,
    )
    scatterplot_matrix(
        dataframe,
        use_param_index,
        file=output_dir + "/sp_matrix.png",
        kind="scatter",
        distinguish_param_setups=False,
    )
    scatterplot_matrix(
        dataframe,
        use_param_index,
        file=output_dir + "/sp_matrix_dist_reg.png",
        kind="reg",
        distinguish_param_setups=True,
    )
    scatterplot_matrix(
        dataframe,
        use_param_index,
        file=output_dir + "/sp_matrix_dist.png",
        kind="scatter",
        distinguish_param_setups=True,
    )

    # radar plots
    radar_plot(
        dataframe, file=output_dir + "/radar_dist.png", distinguish_hyperparam=True
    )
    radar_plot(dataframe, file=output_dir + "/radar.png", distinguish_hyperparam=False)

    # scatter plots for parirs of objectives
    os.makedirs(output_dir + "/scatterpl", exist_ok=True)
    for i, obj_i in enumerate(obj):
        for j in range(i + 1, len(obj)):
            try:
                scatterplot(
                    dataframe,
                    [obj_i, obj[j]],
                    file=output_dir + "/scatterpl/" + obj_i + "_" + obj[j] + ".png",
                )
            except IndexError:
                logger = Logger.get_logger()
                logger.warning(
                    f"disabling kde because cov matrix is singular for objectives "
                    f"{obj_i} & {obj[j]}"
                )
                scatterplot(
                    dataframe,
                    [obj_i, obj[j]],
                    file=output_dir + "/scatterpl/" + obj_i + "_" + obj[j] + ".png",
                    kde=False,
                )

    # create plots for the different algortihms
    for algorithm in dataframe[COLNAME_METHOD].unique():
        os.makedirs(output_dir + "/" + str(algorithm), exist_ok=True)
        dataframe_algo = dataframe[dataframe[COLNAME_METHOD] == algorithm]

        # boxplots
        for objective in obj:
            boxplot(
                dataframe_algo,
                objective,
                file=output_dir
                + "/"
                + str(algorithm)
                + "/variational_plots/"
                + objective,
            )

        # scatterplot matrices
        scatterplot_matrix(
            dataframe_algo,
            use_param_index,
            file=output_dir + "/" + str(algorithm) + "/sp_matrix_reg.png",
            kind="reg",
            distinguish_param_setups=False,
        )
        scatterplot_matrix(
            dataframe_algo,
            use_param_index,
            file=output_dir + "/" + str(algorithm) + "/sp_matrix.png",
            kind="scatter",
            distinguish_param_setups=False,
        )
        scatterplot_matrix(
            dataframe_algo,
            use_param_index,
            file=output_dir + "/" + str(algorithm) + "/sp_matrix_dist_reg.png",
            kind="reg",
            distinguish_param_setups=True,
        )
        scatterplot_matrix(
            dataframe_algo,
            use_param_index,
            file=output_dir + "/" + str(algorithm) + "/sp_matrix_dist.png",
            kind="scatter",
            distinguish_param_setups=True,
        )

        # radar plots
        radar_plot(
            dataframe_algo,
            file=output_dir + "/" + str(algorithm) + "/radar_dist.png",
            distinguish_hyperparam=True,
        )
        radar_plot(
            dataframe_algo,
            file=output_dir + "/" + str(algorithm) + "/radar.png",
            distinguish_hyperparam=False,
        )

        # scatter plots for parirs of objectives
        os.makedirs(output_dir + "/" + str(algorithm) + "/scatterpl", exist_ok=True)
        for i, obj_i in enumerate(obj):
            for j in range(i + 1, len(obj)):
                try:
                    scatterplot(
                        dataframe_algo,
                        [obj_i, obj[j]],
                        file=output_dir
                        + "/"
                        + str(algorithm)
                        + "/scatterpl/"
                        + obj_i
                        + "_"
                        + obj[j]
                        + ".png",
                        distinguish_hyperparam=True,
                    )
                except IndexError:
                    logger = Logger.get_logger()
                    logger.warning(
                        f"WARNING: disabling kde because cov matrix is singular "
                        f"for objectives {obj_i} & {obj[j]}"
                    )
                    scatterplot(
                        dataframe_algo,
                        [obj_i, obj[j]],
                        file=output_dir
                        + "/"
                        + str(algorithm)
                        + "/scatterpl/"
                        + obj_i
                        + "_"
                        + obj[j]
                        + ".png",
                        kde=False,
                        distinguish_hyperparam=True,
                    )


def scatterplot_matrix(
    dataframe_in, use_param_index, file=None, kind="reg", distinguish_param_setups=True
):
    """
    dataframe: dataframe containing the data with columns
        [algo, epos, te_d, seed, params, obj1, ..., obj2]
    file: filename to save the plots (if None, the plot will not be saved)
    reg: if True a regression line will be plotted over the data
    distinguish_param_setups: if True the plot will not only distinguish between models,
        but also between the parameter setups
    """
    dataframe = dataframe_in.copy()
    index = list(range(G_DF_PLOT_COL_METRIC_START, dataframe.shape[1]))
    if distinguish_param_setups:
        dataframe_ = dataframe.iloc[:, index]
        dataframe_.insert(
            0,
            "label",
            dataframe[COLNAME_METHOD].astype(str)
            + ", "
            + dataframe[COLNAME_PARAM].astype(str),
        )

        g_p = sns.pairplot(data=dataframe_, hue="label", corner=True, kind=kind)
    else:
        index_ = list(range(G_DF_PLOT_COL_METRIC_START, dataframe.shape[1]))
        index_.insert(0, G_DF_TASK_COL)
        dataframe_ = dataframe.iloc[:, index_]

        g_p = sns.pairplot(data=dataframe_, hue=COLNAME_METHOD, corner=True, kind=kind)

    for i in range(len(index)):
        for j in range(len(index)):
            if i >= j:
                g_p.axes[i, j].set_xlim((-0.1, 1.1))
                for k in range(j):
                    g_p.axes[j, k].set_ylim((-0.1, 1.1))

    g_p.fig.set_size_inches(12.5, 12)
    if use_param_index and distinguish_param_setups:
        sns.move_legend(g_p, loc="upper right", bbox_to_anchor=(1.0, 1.0), ncol=3)
    else:
        sns.move_legend(g_p, loc="upper right", bbox_to_anchor=(1.0, 1.0), ncol=1)
    plt.tight_layout()

    if file is not None:
        plt.savefig(file, dpi=300)


def scatterplot(dataframe_in, obj, file=None, kde=True, distinguish_hyperparam=False):
    """
    dataframe: dataframe containing the data with columns
        [algo, epos, te_d, seed, params, obj1, ..., obj2]
    obj1 & obj2: name of the objectives which shall be plotted against each other
    file: filename to save the plots (if None, the plot will not be saved)
    kde: if True the distribution of the points will be estimated and plotted as kde plot
    distinguish_param_setups: if True the plot will not only distinguish between models,
        but also between the parameter setups
    """
    obj1, obj2 = obj

    dataframe = dataframe_in.copy()
    dataframe[COLNAME_PARAM] = dataframe[COLNAME_PARAM].astype(str)

    if distinguish_hyperparam:
        if kde:
            g_p = sns.jointplot(
                data=dataframe,
                x=obj1,
                y=obj2,
                hue=COLNAME_PARAM,
                xlim=(-0.1, 1.1),
                ylim=(-0.1, 1.1),
                kind="kde",
                zorder=0,
                levels=8,
                alpha=0.35,
                warn_singular=False,
            )
            gg_p = sns.scatterplot(
                data=dataframe, x=obj1, y=obj2, hue=COLNAME_PARAM, ax=g_p.ax_joint
            )
        else:
            g_p = sns.jointplot(
                data=dataframe,
                x=obj1,
                y=obj2,
                hue=COLNAME_PARAM,
                xlim=(-0.1, 1.1),
                ylim=(-0.1, 1.1),
            )
            gg_p = g_p.ax_joint
    else:
        if kde:
            g_p = sns.jointplot(
                data=dataframe,
                x=obj1,
                y=obj2,
                hue=COLNAME_METHOD,
                xlim=(-0.1, 1.1),
                ylim=(-0.1, 1.1),
                kind="kde",
                zorder=0,
                levels=8,
                alpha=0.35,
                warn_singular=False,
            )
            gg_p = sns.scatterplot(
                data=dataframe,
                x=obj1,
                y=obj2,
                hue=COLNAME_METHOD,
                style=COLNAME_PARAM,
                ax=g_p.ax_joint,
            )
        else:
            g_p = sns.jointplot(
                data=dataframe,
                x=obj1,
                y=obj2,
                hue=COLNAME_METHOD,
                xlim=(-0.1, 1.1),
                ylim=(-0.1, 1.1),
            )
            gg_p = sns.scatterplot(
                data=dataframe, x=obj1, y=obj2, style=COLNAME_PARAM, ax=g_p.ax_joint
            )

    gg_p.set_aspect("equal")
    gg_p.legend(fontsize=6, loc="best")

    if file is not None:
        plt.savefig(file, dpi=300)


def max_0_x(x_arg):
    """
    max(0, x_arg)
    """
    return max(0, x_arg)


def radar_plot(dataframe_in, file=None, distinguish_hyperparam=True):
    """
    dataframe_in: dataframe containing the data with columns
        [algo, epos, te_d, seed, params, obj1, ..., obj2]
    file: filename to save the plots (if None, the plot will not be saved)
    distinguish_param_setups: if True the plot will not only distinguish between models,
        but also between the parameter setups
    """
    dataframe = dataframe_in.copy()
    if distinguish_hyperparam:
        dataframe.insert(
            0,
            "label",
            dataframe[COLNAME_METHOD].astype(str)
            + ", "
            + dataframe[COLNAME_PARAM].astype(str),
        )
    else:
        dataframe.insert(0, "label", dataframe[COLNAME_METHOD])
    # we need "G_DF_PLOT_COL_METRIC_START + 1" as we did insert the columns 'label' at index 0
    index = list(range(G_DF_PLOT_COL_METRIC_START + 1, dataframe.shape[1]))
    num_lines = len(dataframe["label"].unique())
    _, axis = plt.subplots(
        figsize=(9, 9 + (0.28 * num_lines)), subplot_kw=dict(polar=True)
    )
    num = 0

    # Split the circle into even parts and save the angles
    # so we know where to put each axis.
    angles = list(
        np.linspace(0, 2 * np.pi, len(dataframe.columns[index]), endpoint=False)
    )
    for algo_name in dataframe["label"].unique():
        mean = (
            dataframe.loc[dataframe["label"] == algo_name]
            .iloc[:, index]
            .mean()
            .to_list()
        )
        std = (
            dataframe.loc[dataframe["label"] == algo_name]
            .iloc[:, index]
            .std()
            .to_list()
        )

        angles_ = angles
        # The plot is a circle, so we need to "complete the loop"
        # and append the start value to the end.
        mean = np.array(mean + mean[:1])
        std = np.array(std + std[:1])
        angles_ = np.array(angles_ + angles_[:1])

        # Draw the outline of the data.
        axis.plot(
            angles_,
            mean,
            color=list(plt.rcParams["axes.prop_cycle"])[num]["color"],
            linewidth=2,
            label=algo_name,
        )

        # Fill it in.
        axis.fill_between(
            angles_,
            list(map(max_0_x, mean - std)),
            y2=mean + std,
            color=list(plt.rcParams["axes.prop_cycle"])[num]["color"],
            alpha=0.1,
        )
        num += 1
        num = num % len(list(plt.rcParams["axes.prop_cycle"]))

    # Fix axis to go in the right order and start at 12 o'clock.
    axis.set_theta_offset(np.pi / 2)
    axis.set_theta_direction(-1)

    # Draw axis lines for each angle and label.
    axis.set_thetagrids(np.degrees(angles), dataframe.columns[index])

    axis.set_ylim((0, 1))

    plt.legend(loc="lower right", bbox_to_anchor=(1.0, 1.035), ncol=1, fontsize=10)

    if file is not None:
        plt.savefig(file, dpi=300)


def boxplot(dataframe_in, obj, file=None):
    """
    generate the boxplots
    dataframe_in: dataframe containing the data with columns
        [param_idx, task , algo, epos, te_d, seed, params, obj1, ..., obj2]
    obj: objective to be considered in the plot (needs to be contained in dataframe_in)
    file: foldername to save the plots (if None, the plot will not be saved)
    """
    boxplot_stochastic(dataframe_in, obj, file=file)
    boxplot_systematic(dataframe_in, obj, file=file)


def boxplot_stochastic(dataframe_in, obj, file=None):
    """
    generate boxplot for stochastic variation
    dataframe_in: dataframe containing the data with columns
        [param_idx, task , algo, epos, te_d, seed, params, obj1, ..., obj2]
    obj: objective to be considered in the plot (needs to be contained in dataframe_in)
    file: foldername to save the plots (if None, the plot will not be saved)
    """
    dataframe = dataframe_in.copy()
    os.makedirs(file, exist_ok=True)

    ### stochastic variation
    _, axes = plt.subplots(
        1,
        len(dataframe[COLNAME_METHOD].unique()),
        sharey=True,
        figsize=(3 * len(dataframe[COLNAME_METHOD].unique()), 6),
    )
    # iterate over all algorithms
    for num, algo in enumerate(list(dataframe[COLNAME_METHOD].unique())):
        # distinguish if the algorithm does only have one param setup or multiple
        if len(dataframe[COLNAME_METHOD].unique()) > 1:
            # generate boxplot and swarmplot
            sns.boxplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_IDX_PARAM,
                y=obj,
                ax=axes[num],
                showfliers=False,
                boxprops={"facecolor": (0.4, 0.6, 0.8, 0.5)},
            )
            sns.swarmplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_IDX_PARAM,
                y=obj,
                legend=False,
                ax=axes[num],
            )
            # remove legend, set ylim, set x-label and remove y-label
            axes[num].legend([], [], frameon=False)
            axes[num].set_ylim([-0.1, 1.1])
            axes[num].set_xlabel(algo)
            if num != 0:
                axes[num].set_ylabel("")
        else:
            sns.boxplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_IDX_PARAM,
                y=obj,
                ax=axes,
                showfliers=False,
                boxprops={"facecolor": (0.4, 0.6, 0.8, 0.5)},
            )
            sns.swarmplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_IDX_PARAM,
                y=obj,
                hue=COLNAME_IDX_PARAM,
                legend=False,
                ax=axes,
                palette=sns.cubehelix_palette(
                    n_colors=len(
                        dataframe[dataframe[COLNAME_METHOD] == algo][
                            COLNAME_IDX_PARAM
                        ].unique()
                    )
                ),
            )
            axes.legend([], [], frameon=False)
            axes.set_ylim([-0.1, 1.1])
            axes.set_xlabel(algo)
        plt.tight_layout()
        if file is not None:
            plt.savefig(file + "/stochastic_variation.png", dpi=300)


def boxplot_systematic(dataframe_in, obj, file=None):
    """
    generate boxplot for ssystemtic variation
    dataframe_in: dataframe containing the data with columns
        [param_idx, task , algo, epos, te_d, seed, params, obj1, ..., obj2]
    obj: objective to be considered in the plot (needs to be contained in dataframe_in)
    file: foldername to save the plots (if None, the plot will not be saved)
    """
    dataframe = dataframe_in.copy()
    os.makedirs(file, exist_ok=True)

    ### systematic variation
    _, axes = plt.subplots(
        1,
        len(dataframe[COLNAME_METHOD].unique()),
        sharey=True,
        figsize=(3 * len(dataframe[COLNAME_METHOD].unique()), 6),
    )

    for num, algo in enumerate(list(dataframe[COLNAME_METHOD].unique())):
        # distinguish if the algorithm does only have one param setup or multiple
        if len(dataframe[COLNAME_METHOD].unique()) > 1:
            # generate boxplot and swarmplot
            sns.boxplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_METHOD,
                y=obj,
                ax=axes[num],
                showfliers=False,
                boxprops={"facecolor": (0.4, 0.6, 0.8, 0.5)},
            )
            sns.swarmplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_METHOD,
                y=obj,
                hue=COLNAME_IDX_PARAM,
                legend=False,
                ax=axes[num],
                palette=sns.cubehelix_palette(
                    n_colors=len(
                        dataframe[dataframe[COLNAME_METHOD] == algo][
                            COLNAME_IDX_PARAM
                        ].unique()
                    )
                ),
            )
            # remove legend, set ylim, set x-label and remove y-label
            axes[num].legend([], [], frameon=False)
            axes[num].set_ylim([-0.1, 1.1])
            axes[num].set_xlabel(" ")
            if num != 0:
                axes[num].set_ylabel("")
        else:
            sns.boxplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_METHOD,
                y=obj,
                ax=axes,
                showfliers=False,
                boxprops={"facecolor": (0.4, 0.6, 0.8, 0.5)},
            )
            sns.swarmplot(
                data=dataframe[dataframe[COLNAME_METHOD] == algo],
                x=COLNAME_METHOD,
                y=obj,
                hue=COLNAME_IDX_PARAM,
                legend=False,
                ax=axes,
                palette=sns.cubehelix_palette(
                    n_colors=len(
                        dataframe[dataframe[COLNAME_METHOD] == algo][
                            COLNAME_IDX_PARAM
                        ].unique()
                    )
                ),
            )
            axes.legend([], [], frameon=False)
            axes.set_ylim([-0.1, 1.1])
            axes.set_xlabel(" ")
        plt.tight_layout()

    if file is not None:
        plt.savefig(file + "/systematic_variation.png", dpi=300)



"""
select trainer
"""
from domainlab.algos.trainers.train_basic import TrainerBasic
from domainlab.algos.trainers.train_dial import TrainerDIAL
from domainlab.algos.trainers.train_hyper_scheduler import TrainerHyperScheduler
from domainlab.algos.trainers.train_matchdg import TrainerMatchDG
from domainlab.algos.trainers.train_mldg import TrainerMLDG


class TrainerChainNodeGetter(object):
    """
    Chain of Responsibility: node is named in pattern Trainer[XXX] where the string
    after 'Trainer' is the name to be passed to args.trainer.
    """

    def __init__(self, str_trainer):
        """__init__.
        :param args: command line arguments
        """
        self._list_str_trainer = None
        if str_trainer is not None:
            self._list_str_trainer = str_trainer.split("_")
            self.request = self._list_str_trainer.pop(0)
        else:
            self.request = str_trainer

    def __call__(self, lst_candidates=None, default=None, lst_excludes=None):
        """
        1. construct the chain, filter out responsible node,
        create heavy-weight business object
        2. hard code seems to be the best solution
        """
        if lst_candidates is not None and self.request not in lst_candidates:
            raise RuntimeError(
                f"desired {self.request} is not supported \
                               among {lst_candidates}"
            )
        if default is not None and self.request is None:
            self.request = default
        if lst_excludes is not None and self.request in lst_excludes:
            raise RuntimeError(
                f"desired {self.request} is not supported among {lst_excludes}"
            )

        chain = TrainerBasic(None)
        chain = TrainerDIAL(chain)
        chain = TrainerMatchDG(chain)
        chain = TrainerMLDG(chain)
        chain = TrainerHyperScheduler(chain)
        node = chain.handle(self.request)
        head = node
        while self._list_str_trainer:
            self.request = self._list_str_trainer.pop(0)
            node2decorate = self.__call__(lst_candidates, default, lst_excludes)
            head.extend(node2decorate)
            head = node2decorate
        return node
